#include <allegro.h>
#include <stdio.h>

extern int read_midi (char *name, char **title);
extern char const *GM_names[];

static int BACK, TEXT, MUSIC, SHADOW;
static int WHITE[18], BLACK[18];
static int W = 4, H = 21; /* size of a white key */
static int YGAP; /* space between rows */
static int LEFTX = 6, TOPY, BOTTOMY, COLUMNS = 11;
static int KEY_POS[] = { 0, 0, 1, 1, 2,	3, 3, 4, 4, 5, 5, 6};
static int program[16];
static int argc;
static char **argv;
static int file_num;
static int midi_len;
static char *title;
static int scroll_pos, scroll_lines;

static void
draw_black_key (int x, int y, int s)
{
	int c = BLACK[s];
	int w = W / 3;
	int h = 2 * H / 3;
	rectfill (screen, x + W - w, y, x + W + w - 1, y + h - 1, c);
}

static void
draw_white_sharp_key (int x, int y, int s)
{
	int c = WHITE[s];
	rectfill (screen, x, y, x + W - W / 3 - 1, y + 2 * H / 3 - 1, c);
	rectfill (screen, x, y + 2 * H / 3, x + W - 1, y + H - 1, c);
}

static void
draw_white_flat_key (int x, int y, int s)
{
	int c = WHITE[s];
	rectfill (screen, x + W / 3, y, x + W - 1, y + 2 * H / 3 - 1, c);
	rectfill (screen, x,  y + 2 * H / 3, x + W - 1, y + H - 1, c);
}

static void
draw_white_key (int x, int y, int s)
{
	int c = WHITE[s];
	rectfill (screen, x + W / 3, y, x + W - W / 3 - 1, y + 2 * H / 3 - 1, c);
	rectfill (screen, x, y + 2 * H / 3, x + W - 1, y + H - 1, c);
}

static void
draw_key (int c, int note, int s)
{
	int x = LEFTX + (c & 1) * SCREEN_W / 2;
	int y = TOPY + (c / 2) * (H + YGAP);
	int octave = note / 12;
	int n = note % 12;
	x += ((octave % COLUMNS) * 7 + KEY_POS[n]) * W;
	y += (10 / COLUMNS) * H - (octave / COLUMNS) * H;
	
	if (n == 1 || n == 3 || n == 6 || n == 8 || n == 10)
		draw_black_key (x, y, s);
	if (n == 0 || n == 5)
		draw_white_sharp_key (x, y, s);
	if (n == 4 || n == 11)
		draw_white_flat_key (x, y, s);
	if (n == 2 || n == 7 || n == 9)
		draw_white_key (x, y, s);
	
}

static void
draw_progress_bar (int pos, int len, int cl)
{
	int l = LEFTX;
	int t = SCREEN_H - YGAP * 2;
	int r = SCREEN_W - LEFTX - 1;
	int b = SCREEN_H - YGAP - 1;
	int p = l + (r - l - 11) * pos / len;
	static int last_p, otl = 0;
	static BITMAP *bar = NULL;
	if (!bar)
	{
		bar = create_bitmap (SCREEN_W, YGAP);
	}
	static char str[256];
	sprintf (str, "%ld / %d", midi_pos, len);
	int ctl = text_length (font, str);
	int tl;
	tl = MAX (ctl, otl);
	otl = ctl;
	set_clip (bar, SCREEN_W / 2 - tl / 2, 0,
		SCREEN_W / 2 - tl / 2 + tl, bar->h - 1);
	rectfill (bar, p + 2, 0, r, bar->h - 1, WHITE[16]);
	rectfill (bar, l, 0, p - 1, bar->h - 1, SHADOW);
	rectfill (bar, p, 0, p + 1, bar->h - 1, MUSIC);
	textout (bar, font, str, SCREEN_W / 2 - ctl / 2, 0, BACK);
        acquire_screen ();
	rectfill (screen, p, t, p + 1, b, MUSIC);
	if (cl)
	{
		rectfill (screen, p + 2, t, r, b, WHITE[16]);
		rectfill (screen, l, t, p - 1, b, SHADOW);
	}
	else
	{
		if (p > last_p)
			rectfill (screen, last_p - 1, t, p - 1, b, SHADOW);
		else
			rectfill (screen, p + 2, t, last_p + 2, b, WHITE[16]);
	}
	blit (bar, screen, SCREEN_W / 2 - tl / 2, 0,
		SCREEN_W / 2 - tl / 2, t, tl, bar->h);
		
	release_screen ();
	last_p = p;
}

static void
draw_info (void)
{
	int x = SCREEN_W / 2;
	textprintf_centre (screen, font, x, YGAP * 0, TEXT,
		"Driver: %s / %s", midi_driver->name, digi_driver->name);
	textprintf_centre (screen, font, x, YGAP * 1, TEXT,
		"Playing file %d/%d: %s", file_num, argc - 1,
		get_filename (argv[file_num]));
	DIALOG d = {d_textbox_proc, LEFTX, YGAP * 2, SCREEN_W - LEFTX * 2,
		YGAP + 8, TEXT, BACK, 0, 0, 0, scroll_pos, title, NULL, NULL};
	d_textbox_proc (MSG_DRAW, &d, 0);
	scroll_lines = d.d1;
}

static void
draw_program (int c, int p)
{
	char const *str;
	if (c == 9)
	{
		str = "Percussion";
		program[c] = 17;
	}
	else
	{
		if (p == -1)
			p = 0;
		program[c] = p;
		str	= GM_names[p];
	}

	int r, g, b;
	hsv_to_rgb (360 * program[c] / 128.0, 1, 0.5, &r, &g, &b);
	WHITE[c] = makecol (r, g, b);
	hsv_to_rgb (360 * program[c] / 128.0, 1, 1, &r, &g, &b);
	BLACK[c] = makecol (r, g, b);

	int l = LEFTX + (c & 1) * SCREEN_W / 2;
	int t = TOPY + (c / 2) * (H + YGAP);
	int th = text_height (font);

	rectfill (screen, l, t - th, l + COLUMNS * 7 * W - 1, t - 1, BACK);
	textprintf (screen, font, l - 1, t - th, SHADOW, "%s", str);
	textprintf (screen, font, l, t - th - 1, WHITE[c], "%s", str);
}

static void
midikey (int msg, int byte1, int byte2)
{
	int c = 0;
	int n = -1;
	int p = -1;
	int v = 0;

	if (msg >= 128 && msg < 128 + 16)
	{
		c = msg - 128;
		n = byte1;
		v = 0;
	}
	else if (msg >= 144 && msg < 144 + 16)
	{
		c = msg - 144;
		n = byte1;
		v = byte2;
	}
	else if (msg >= 192 && msg < 192 + 16)
	{
		c = msg - 192;
		p = byte1;
	}
	else
	{
		return;
	}

        acquire_screen();
	if (p >= 0 || program[c] == -1)
	{
		draw_program (c, p);
	}
		

	if (n >= 0)
	{
		draw_key (c, n, v ? c : 16);
	}
        release_screen();

}

static void
redraw (void)
{
	int c;

	for (c = 0; c < 16; c++)
	{
		int n;
		for (n = 0; n < 132; n++)
		{
			draw_key (c, n, 17);
		}
	}

	draw_info ();
	
	draw_progress_bar (midi_pos, midi_len, 1);
}

static void
colors (void)
{
	BACK = makecol (200, 200, 200);
	SHADOW = makecol (150, 150, 150);
	TEXT = makecol (0, 0, 0);
	MUSIC = makecol (255, 0, 255);

	WHITE[16] = makecol (255, 255, 255);
	BLACK[16] = makecol (0, 0, 0);
	WHITE[17] = makecol (220, 220, 220);
	BLACK[17] = makecol (100, 100, 100);
}

static void
reset (void)
{
	text_mode (-1);
	clear_to_color (screen, BACK);
	
	scroll_pos = 0;
		
	int i;
	for (i = 0; i < 16; i++)
		program[i] = -1;
}

static void
close_button (void)
{
	simulate_keypress (27 + (KEY_ESC << 8));
}

extern struct FONT_VTABLE *font_vtable_mono;
extern DATAFILE data[];

extern int _mangled_main (int argc_, char *argv_[]);
int
main (int argc_, char *argv_[])
{
	if (allegro_init () != 0)
		return 1;

	argc = argc_;
	argv = argv_;

	((FONT *)data[0].dat)->vtable = font->vtable;;
	font = data[0].dat;

	if (argc < 2)
	{
		allegro_message ("Usage: 'almid midi1 midi2 ...'\n");
		return 1;
	}
	
	YGAP = text_height (font) + 2;
	TOPY = YGAP * 6;
	BOTTOMY = YGAP * 3;
	set_color_depth (desktop_color_depth ());

	if (set_gfx_mode (GFX_AUTODETECT_WINDOWED, 640,
		TOPY + (H + YGAP) * 8 + BOTTOMY, 0, 0) != 0)
	{
		allegro_message
			("Unable to open display window\n%s\n", allegro_error);
		return 1;
	}

	colors ();

	install_keyboard ();
	install_timer ();
        install_mouse();

    if (set_display_switch_mode(SWITCH_BACKGROUND))
		set_display_switch_mode(SWITCH_BACKAMNESIA);

    enable_hardware_cursor();
    select_mouse_cursor(MOUSE_CURSOR_ARROW);
	
	#if (ALLEGRO_MINOR == 0)
		set_window_close_button (1);
		set_window_close_hook (close_button);
	#else
		set_close_button_callback (close_button);
	#endif

	/* install a MIDI sound driver */
	set_volume_per_voice (2);
	if (install_sound (DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) != 0)
	{
		allegro_message	("Error initialising sound\n%s\n", allegro_error);
		return 1;
	}

        show_mouse(screen);
	file_num = 1;
	while (1)
	{
		MIDI *the_music;

		reset ();

		/* read in the MIDI file */
		the_music = load_midi (argv[file_num]);
		if (the_music)
		{
			read_midi (argv[file_num], &title);
			/* start up the MIDI file */
			play_midi (the_music, FALSE);
			while (midi_pos < 0);
			midi_seek (INT_MAX);
			midi_len = -midi_pos;
			stop_midi ();

			midi_msg_callback = midikey;
			redraw ();

			play_midi (the_music, FALSE);

			while (midi_pos < 0);
go_on:
			while (!keypressed () && midi_pos >= 0)
			{
#if (defined (ALLEGRO_LINUX) && ALLEGRO_SUB_VERSION == 0)
				struct timeval tv = {0,	1};
				select (0, NULL, NULL, NULL, &tv);
#elif (defined (ALLEGRO_WINDOWS) && ALLEGRO_SUB_VERSION == 0)
				Sleep (1);
#else
				rest (1);
#endif

				draw_progress_bar (midi_pos, midi_len, 0);
			}

		}
		else
		{
			allegro_message ("Error reading MIDI file '%s'\n", argv[file_num]);
		}

		if (keypressed ())
		{
			int k = readkey () >> 8;
			if (k == KEY_PGUP)
			{
				if (scroll_pos > 0)
				{
					scroll_pos--;
					draw_info ();
				}
				goto go_on;
			}
			else if (k == KEY_PGDN)
			{
				if (scroll_pos < scroll_lines - 1)
				{
					scroll_pos++;
					draw_info ();
				}
				goto go_on;
			}
			else if (k == KEY_RIGHT)
			{
				midi_seek (midi_pos + 4);
				goto go_on;
			}
			else if (k == KEY_LEFT)
			{
				redraw ();
				midi_seek (midi_pos - 4);
				goto go_on;
			}
			if (k == KEY_UP)
			{
				midi_seek (midi_pos + midi_len / 10);
				goto go_on;
			}
			else if (k == KEY_DOWN)
			{
				redraw ();
				midi_seek (midi_pos - midi_len / 10);
				goto go_on;
			}
			else if (k == KEY_SPACE)
			{
				file_num++;
				if (file_num == argc)
					file_num = 1;

			}
			else if (k == KEY_BACKSPACE)
			{
				file_num--;
				if (file_num == 0)
					file_num = argc - 1;
			}
			else if (k == KEY_ESC)
				break;
			else
				goto go_on;
		}
		else
		{
			file_num++;
			if (file_num == argc)
				file_num = 1;
		}

		/* destroy the MIDI file */
		midi_msg_callback = NULL;
		destroy_midi (the_music);
	}

	return 0;
}

END_OF_MAIN ()
