#include <allegro.h>
#include <string.h>

int read_midi (char *name, char **title);
int
read_midi (char *name, char **title)
{
	int M, T, h, d, r, k;
	int hl, t, tn;
	int div0, div1, div2;
	int com = 0;
	int pos;
	int it;
	PACKFILE *notesfile;
	int len;
	int i;

	*title = NULL;

	if (!(notesfile = pack_fopen (name, "r")))
	{
		return -1;
	}

	M = pack_getc (notesfile);
	T = pack_getc (notesfile);
	h = pack_getc (notesfile);
	d = pack_getc (notesfile);
	hl = pack_getc (notesfile) << 24;
	hl += pack_getc (notesfile) << 16;
	hl += pack_getc (notesfile) << 8;
	hl += pack_getc (notesfile);
	t = pack_getc (notesfile) << 8;
	t += pack_getc (notesfile);
	tn = pack_getc (notesfile) << 8;
	tn += pack_getc (notesfile);
	/* how much MIDI time makes up one quarter note */
	div1 = pack_getc (notesfile);
	div2 = pack_getc (notesfile);
	/* Alternatively, if the first byte has highest bit set, the time is given
	 * in another format: The first byte is the negative FPS, normally -24, -25,
	 * -29 or -30; the second byte then is the sub-resolution, usually 4, 8, 10,
	 * 80 or 1000. */
	if (div1 < 128) /* i.e., first byte is positive */
	{
                /* assume one second = quarter note */
		div0 = div1 * 256 + div2;
	}
	else
	{
		div0 = (div1 - 256) * div2;
	}

	if (M != 'M' || T != 'T' || h != 'h' || d != 'd' || hl != 6)
	{
		return -1;
	}

	if (t != 0 && t != 1)
	{
		return - 1;
	}

	if (t == 0)
		tn = 1;

	/* Read all tracks. */
	for (t = 0; t < tn; t++)
	{
		pos = 0;
		M = pack_getc (notesfile);
		T = pack_getc (notesfile);
		r = pack_getc (notesfile);
		k = pack_getc (notesfile);
		if (M != 'M' || T != 'T' || r != 'r' || k != 'k')
		{
			return -1;
		}

		len = pack_getc (notesfile) << 24;
		len += pack_getc (notesfile) << 16;
		len += pack_getc (notesfile) << 8;
		len += pack_getc (notesfile);

		/* Read the track. */
		for (it = 0; it < len; it++)
		{
			int c, tt, next;
			int vl;
			if (pack_feof (notesfile))
				break;

			tt = 0;
			for (vl = 0; vl < 4; vl++)
			{
				int b;
				b = pack_getc (notesfile);
				tt += b & 127;
				if (b > 127)
					tt *= 128;
				else
					break;
			}
			pos += tt;

			c = pack_getc (notesfile);
			if (c & 0x80)
			{
				com = c;
				next = pack_getc (notesfile);
			}
			else	/* running status */
			{
				next = c;
				c = com;
			}

			if (c >= 0x80 && c <= 0x8f)	/* note off */
			{
				pack_getc (notesfile);
			}
			else if (c >= 0x90 && c <= 0x9f)	/* note on */
			{
				pack_getc (notesfile);
			}
			else if (c >= 0xa0 && c <= 0xaf)	/* aftertouch */
			{
				pack_getc (notesfile);
			}
			else if (c >= 0xb0 && c <= 0xbf)	/* controller */
			{
				pack_getc (notesfile);
			}
			else if (c >= 0xc0 && c <= 0xcf)	/* program change */
			{
				
			}
			else if (c >= 0xd0 && c <= 0xdf)	/* channel pressure */
			{

			}
			else if (c >= 0xe0 && c <= 0xef)	/* pitch wheel */
			{
				pack_getc (notesfile);
			}
			else if (c == 0xf0 || c == 0xf7)	/* sysex */
			{
				int l;

				l = 0;
				for (vl = 0; vl < 4; vl++)
				{
					int b;
					b = vl ? pack_getc (notesfile) : next;
					l += b & 127;
					if (b > 127)
						l *= 128;
					else
						break;
				}

				for (i = 0; i < l; i++)
				{
					pack_getc (notesfile);
				}
			}
			else if (c == 0xff)	/* meta */
			{
				int t2, l;
				t2 = next;

				l = 0;
				for (vl = 0; vl < 4; vl++)
				{
					int b;
					b = pack_getc (notesfile);
					l += (b & 127);
					if (b > 127)
						l *= 128;
					else
						break;
				}

				if (t2 == 0x2f)
					break;
				else if (t2 >= 1 && t2 <= 9)
				{
					int title_len = *title ? strlen (*title) : 0;
					if (title_len)
						title_len++;
					*title = realloc (*title, title_len + l + 1);
					if (title_len)
					{
						(*title)[title_len - 1] = ' ';
					}
					for (i = 0; i < l; i++)
					{
						int c2 = pack_getc (notesfile);
						(*title)[title_len + i] = c2;
					}
					(*title)[title_len + i] = '\0';
				}
				else
				{
					for (i = 0; i < l; i++)
					{
						pack_getc (notesfile);
					}

				}

			}
		}
	}
	pack_fclose (notesfile);
	return 0;
}
